(***********************************************************)
(*  This file is distributed under the terms of the        *)
(*    GNU Lesser General Public License version 3          *)
(*                                                         *)
(*          Copyright 2010: Damien Pous.                   *)
(*                                                         *)
(***********************************************************)

(* functions composition *)
val ( >> ) : ('a -> 'b) -> ('b -> 'c) -> 'a -> 'c

(* formula constructors (par and bottom are handled as sequents) *)
type 'a constr = [
  | `Var of int
  | `Rav of int
  | `One 
  | `Ten of 'a * 'a
  | `Opl of 'a * 'a
  | `Wit of 'a * 'a
]

(* formulae *)
type t = [t constr | `Seq of seq]

(* sequents *)
and seq = node Seq.t

(* nodes *)
and node = t constr



(* formula constructions *)
val var: int -> t
val rav: int -> t
val one: t
val bot: t
val ten: t -> t -> t
val par: t -> t -> t
val wit: t -> t -> t
val opl: t -> t -> t

(* sequent constructions *)
val empty: seq
val singleton: t -> seq
val append: seq -> seq -> seq
val cons: t -> seq -> seq
val snoc: seq -> t -> seq

(* building a sequent out of a list *)
val seq_of_list: t list -> seq

(* dual of a formula *)
val dual: t -> t
val dual_seq: seq -> t

(* print a formula / sequent *)
val print: out_channel -> t -> unit
val print_seq: out_channel -> seq -> unit

(* size of a formula / sequent (number of leaves) *)
val size: t -> int
val size_seq: seq -> int

(* normalisation function for sequents:
   - equates sequents that are equal up to injective renaming of atoms
   - equates sequents that are equal modulo cyclic permutation
   (of course, this function has a bad complexity)
 *)
val normalise: seq -> seq

(*
(* equality test *)
val equal: t -> t -> bool
val equal_seq: seq -> seq -> bool

(* hashtable over sequents *)
module Table: Hashtbl.S with type key = seq
*)

(* annotated version *)
module A: sig

  (* formulae *)
  type 'i t = [('i t * 'i) constr | `Seq of 'i seq]
  
  (* sequents *)
  and 'i seq = ('i node * 'i) Seq.t
  
  (* nodes *)
  and 'i node = ('i t * 'i) constr
  
  (* sequent operations *)
  val singleton: 'i t * 'i -> 'i seq
  val cons: 'i t * 'i -> 'i seq -> 'i seq
  val snoc: 'i seq -> 'i t * 'i -> 'i seq
  
  (* size of a formula / sequent (number of leaves) *)
  val size: 'i t -> int
  val size_seq: 'i seq -> int
  
end

(* removing annotations *)
val clear: 'i A.t -> t
val clear_seq: 'i A.seq -> seq
