(***********************************************************)
(*  This file is distributed under the terms of the        *)
(*    GNU Lesser General Public License version 3          *)
(*                                                         *)
(*          Copyright 2010: Damien Pous.                   *)
(*                                                         *)
(***********************************************************)

open Formula

(* synchronous nodes / atoms *)
type 'a sync = [
  | `Var of int
  | `Rav of int
  | `One 
  | `Ten of 'a * 'a
  | `Opl of 'a * 'a
]

(* sequents of synchronous nodes *)
type sseq = t sync Seq.t

(* search algorithm *)
let run (heuristic: seq -> bool): seq -> bool =
  
  (* asynchronous phase on l++k , where l contains only atoms and synchronous formulae *)
  let rec async (l: sseq) (k: seq) =
    if Seq.is_empty k then 
      heuristic (l:>seq) && Seq.cyclic_positions focus l
    else let a,k = Seq.head k, Seq.tail k in 
      match a with
        | `Wit(a,b) -> async l (cons a k) && async l (cons b k)
        | #sync as a -> async (Seq.snoc l a) k     (* update color graph here *)
  
  (* focused phase on x::l *)
  and focused (x: t) (l: sseq) = 
    match x with
      | `Ten(a,b) -> Seq.splittings (fun h k -> focused b h && focused a k) l (* split color graph here *)
      | `Opl(a,b) -> focused a l || focused b l
      | `One -> Seq.is_empty l
      | `Rav i -> Seq.is_singleton (`Var i) l
      | `Var _ 
      | `Wit _
      | `Seq _ -> async l (singleton x)
  
  (* attempt to focus on x in x::l *)
  and focus (x: t sync) (l: sseq) =
    match x with
      | `Var _ -> false
      | _ -> focused (x:>t) l
  
in async Seq.empty

module A = struct

open A
  
(* synchronous nodes *)
type 'i snode = ('i t * 'i) sync
  
(* sequents of synchronous nodes *)
and 'i sseq = ('i snode * 'i) Seq.t

(* is a sequent a given singleton *)
let is_singleton x l =
  not (Seq.is_empty l) && Seq.is_empty (Seq.tail l) && fst (Seq.head l) = x


(* run focused proof search with a given heuristic *)
let run (heuristic: 'i seq -> bool): 'i seq -> bool =
  
  (* asynchronous phase on l++k , where l contains only atoms and synchronous formulae *)
  let rec async (l: 'i sseq) (k: 'i seq) =
    if Seq.is_empty k then 
      heuristic (l:>_ seq) && Seq.cyclic_positions focus l
    else let a,k = Seq.head k, Seq.tail k in 
      match a with
        | `Wit(a,b), _ -> async l (cons a k) && async l (cons b k)
        | #sync,_ as a -> async (Seq.snoc l a) k
  
  (* focused phase on x::l *)
  and focused (x: 'i t * 'i) (l: 'i sseq) = 
    match fst x with
      | `Ten(a,b) -> Seq.splittings (fun h k -> focused b h && focused a k) l
      | `Opl(a,b) -> focused a l || focused b l
      | `One -> Seq.is_empty l
      | `Rav i -> is_singleton (`Var i) l
      | `Var _ 
      | `Wit _
      | `Seq _ -> async l (singleton x)
  
  (* attempt to focus on x in x::l *)
  and focus (x: 'i snode * 'i) (l: 'i sseq) =
    match fst x with
      | `Var _ -> false
      | _ -> focused (x:>'i t * 'i) l

  in
    async Seq.empty

end
