(***********************************************************)
(*  This file is distributed under the terms of the        *)
(*    GNU Lesser General Public License version 3          *)
(*                                                         *)
(*          Copyright 2010: Damien Pous.                   *)
(*                                                         *)
(***********************************************************)

(* fifo persistent queues *)
module Queue : sig 
  type +'a t
  val empty: 'a t
  val is_empty: 'a t -> bool
  val snoc: 'a t -> 'a -> 'a t
  val head: 'a t -> 'a 
  val tail: 'a t -> 'a t
  val fold: ('b -> 'a -> 'b) -> 'b -> 'a t -> 'b
end = struct
  type 'a t = 'a list * 'a list
  let empty = [],[]
  let is_empty (f,r) = f=[]
  let checkf f r = if f=[] then List.rev r,[] else f,r
  let snoc (f,r) x = checkf f (x::r)
  let head = function
    | [],_ -> failwith "Queue.head: empty"
    | x::_,_ -> x
  let tail = function
    | [],_ -> failwith "Queue.tail: empty"
    | _::f,r -> checkf f r
  let rec fold_right f b = function
    | [] -> b
    | x::q -> f (fold_right f b q) x
  let fold f b (l,r) = fold_right f (List.fold_left f b r) l
end


(* sequences (cf. Okasaki's book) *)

type 'a t = E | C of 'a * 'a t Lazy.t Queue.t

let empty = E

let is_empty = function E -> true | _ -> false

let link t s = match t with
  | E -> failwith "Seq.link empty"
  | C(x,q) -> C(x,Queue.snoc q s)

let rec link_all q =
  let t = Lazy.force (Queue.head q) in
  let q' = Queue.tail q in
    if Queue.is_empty q' then t else link t (lazy (link_all q'))

let append xs ys = match xs,ys with
  | E,_ -> ys
  | _,E -> xs
  | _ -> link xs (lazy ys)

let head = function
  | E -> failwith "Seq.head: empty"
  | C(x,_) -> x

let tail = function
  | E -> failwith "Seq.tail: empty"
  | C(_,q) -> if Queue.is_empty q then E else link_all q

let rec force = function
  | E -> ()
  | C(x,q) -> Queue.fold (fun () y -> force (Lazy.force y)) () q

let rec fold_right f l b = match l with
  | E -> b
  | C(x,q) -> f x (Queue.fold (fun acc y -> fold_right f (Lazy.force y) acc) b q)

let rec fold_left f b l =
  if is_empty l then b
  else fold_left f (f b (head l)) (tail l)

let singleton x = C(x,Queue.empty)

let cons x xs = append (singleton x) xs

let snoc xs x = append xs (singleton x)

let equal f = 
  let rec equal x y = match x,y with
    | E,E -> true
    | C(x,h),C(y,k) -> f x y && 
	(match Queue.is_empty h, Queue.is_empty k with
	   | true,true -> true
	   | false,false -> equal (link_all h) (link_all k)
	   | _ -> false)
    | _ -> false
  in equal

let cyclic_positions f =
  let rec aux acc = function
    | E -> false
    | C(x,q) -> 
	  let q = if Queue.is_empty q then E else link_all q in
	    f x (append q acc) || aux (snoc acc x) q
  in aux E

let fold_splittings f = 
  let rec aux acc l b = 
    let b = f acc l b in
      match l with
	| E -> b
	| C(x,q) -> 
	    if Queue.is_empty q then aux (snoc acc x) E b
	    else aux (snoc acc x) (link_all q) b
  in aux E
let splittings f l =
  try fold_splittings (fun h k () -> if f h k then raise Not_found) l (); false
  with Not_found -> true

(* conversion between sequences and lists *)
let rec of_list = function
  | [] -> empty 
  | x::q -> cons x (of_list q)
let rec to_list l = if is_empty l then [] else head l :: to_list (tail l)

let rec to_stream l = 
  let c = ref 0 in
  let r = ref l in
    Stream.from 
      (fun i -> assert (!c = i); incr c; 
	 let s = !r in
	   if is_empty s then None
	   else (r := tail s; Some (head s))
      )

let is_singleton x = function
  | E -> false
  | C(y,q) -> x=y && Queue.is_empty q


(*
(* naive implementation of sequences, as lists *)
type 'a t = 'a list
let empty = []
let is_empty l = l=[]
let append = List.append
let head = List.hd
let tail = List.tl
let rec fold f b = function
  | [] -> b
  | x::q -> f (fold f b q) x
let singleton x = [x]
let cons x xs = x::xs
let snoc xs x = append xs (singleton x)
let rec cyclic_positions f =
  let rec aux acc = function
    | [] -> false
    | x::q -> f x (q@acc) || aux (acc@[x]) q
  in aux []
let rec splittings f = 
  let rec aux acc l = 
    f acc l || match l with
	| [] -> false
	| x::q -> aux (acc@[x]) q 
  in aux []
let to_list l = l  
let of_list l = l
*)
